﻿using System;
using System.Text.RegularExpressions;
using System.Web;

namespace Palmmedia.FeedSubscriberCounter
{
    /// <summary>
    /// <see cref="IHttpModule"/> counting the number of readers of a feed.
    /// The statistcs of the last day are reported by an <see cref="IFeedSubcriberNotifier"/>
    /// </summary>
    public class FeedSubscriberCounterModule : IHttpModule
    {
        // TODO: Change to your feed's path
        /// <summary>
        /// The absolute path of your feed.
        /// If your feed is located at 'http://www.mydomain.xyz/Blog/Feed the absolute path' is '/Blog/Feed'
        /// </summary>
        private const string FEEDPATH = "/Blog/Feed";
        
        /// <summary>
        /// The pattern the useragent of a feed aggregators must match.
        /// </summary>
        private const string AGGREGATOR_PATTERN = @".+?(\d*).?(?>subscribers|readers|users).?(\d*).*";

        /// <summary>
        /// Initializes a new instance of the <see cref="FeedSubscriberCounterModule"/> class.
        /// </summary>
        public FeedSubscriberCounterModule()
        {
            // TODO: Depending on how you store your stats, you have to use a singleton implementation of the following objects.
            // This IHttpModule is reinstantiated from time to time, so if you keep your statistics in memory you have to make sure you reuse the same objects
            // This can be done by implementing a singleton or by using a ServiceLayer like Unity
            // But be aware: Even singletons are lost if application is shut down, consider using an IFeedStatisticsStore with a database or file storage
            this.FeedSubscriberNotifier = new EmailFeedSubscriberNotifier();
            this.FeedStatisticsStore = new FileFeedStatisticsStore();
        }

        /// <summary>
        /// Gets or sets the feed subcriber notifier.
        /// </summary>
        public IFeedSubscriberNotifier FeedSubscriberNotifier { get; set; }

        /// <summary>
        /// Gets or sets the feed statistics store.
        /// </summary>
        public IFeedStatisticsStore FeedStatisticsStore { get; set; }

        /// <summary>
        /// Initializes a module and prepares it to handle requests.
        /// </summary>
        /// <param name="context">An <see cref="T:System.Web.HttpApplication"/> that provides access to the methods, properties, and events common to all application objects within an ASP.NET application</param>
        public void Init(HttpApplication context)
        {
            context.BeginRequest += new EventHandler(this.Context_BeginRequest);
        }

        /// <summary>
        /// Handles the BeginRequest event of the <see cref="HttpApplication"/>.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        protected void Context_BeginRequest(object sender, EventArgs e)
        {
            var request = HttpContext.Current.Request;

            if (request.Url.AbsolutePath.Equals(FEEDPATH, StringComparison.OrdinalIgnoreCase))
            {
                // Send statistics if required
                if (this.FeedStatisticsStore.SendReport)
                {
                    var statistics = this.FeedStatisticsStore.GetStatistcsToReport();
                    this.FeedSubscriberNotifier.ReportStatistics(statistics);
                }

                this.RegisterRequest(request);
            }
        }

        /// <summary>
        /// Analyses the request to determinate whether the request's origin is a single user or a feed aggregator.
        /// </summary>
        /// <param name="request">The request.</param>
        private void RegisterRequest(HttpRequest request)
        {
            var match = Regex.Match(request.UserAgent, AGGREGATOR_PATTERN, RegexOptions.Compiled);

            string application = request.Browser.Browser + " " + request.Browser.Version;

            if (application.StartsWith("Unknown", StringComparison.OrdinalIgnoreCase))
            {
                application = Regex.Match(request.UserAgent, @"^(?>\w|-)*", RegexOptions.Compiled).Value;
            }

            if (match.Success)
            {
                string numberOfSubscribers = match.Groups[1].Value;

                if (string.IsNullOrEmpty(numberOfSubscribers))
                {
                    numberOfSubscribers = match.Groups[2].Value;
                }

                this.FeedStatisticsStore.AddFeedAggregator(
                    application, 
                    int.Parse(numberOfSubscribers), 
                    DateTime.Now);
            }
            else
            {   
                // TODO: Calculate HashCode instead of using the full identifier
                string identifier = request.UserHostAddress + request.UserAgent;
                this.FeedStatisticsStore.AddSingleUser(application, identifier, DateTime.Now);
            }
        }

        /// <summary>
        /// Disposes of the resources (other than memory) used by the module that implements <see cref="T:System.Web.IHttpModule"/>.
        /// </summary>
        public void Dispose()
        {
        }
    }
}
