﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Palmmedia.FeedSubscriberCounter
{
    /// <summary>
    /// Stores feed statistics in a file and caches them in memory.
    /// The used diectionaires are cleared as soon as the report is generated.
    /// </summary>
    public class FileFeedStatisticsStore : IFeedStatisticsStore
    {
        // TODO: Change to your desired path or read from a config file
        /// <summary>
        /// The path of the file where statistics should be saved.
        /// </summary>
        private readonly string filename = Path.Combine(HttpContext.Current.Request.PhysicalApplicationPath, "App_Data/feedstatistics.xml");

        /// <summary>
        /// Lock object for thread safety.
        /// </summary>
        private readonly object lockObject = new object();

        /// <summary>
        /// Dictionary containing the application by the identifier of an user.
        /// </summary>
        private Dictionary<string, string> singleUsers = new Dictionary<string, string>();

        /// <summary>
        /// Dictionary containing the number of subsriber by the aggregator.
        /// </summary>
        private Dictionary<string, int> feedAggregators = new Dictionary<string, int>();

        /// <summary>
        /// The day whose statics are currently collected.
        /// </summary>
        private int currentDay;

        /// <summary>
        /// Initializes a new instance of the <see cref="FileFeedStatisticsStore"/> class.
        /// </summary>
        public FileFeedStatisticsStore()
        {
            this.ReadFromFile();
        }

        /// <summary>
        /// Gets a value indicating whether a report should be send or not.
        /// </summary>
        /// <value></value>
        public bool SendReport
        {
            get
            {
                return this.currentDay != DateTime.Now.Day;
            }
        }

        /// <summary>
        /// Adds a single user.
        /// </summary>
        /// <param name="application">The application.</param>
        /// <param name="identfier">The identfier.</param>
        /// <param name="time">The time.</param>
        public void AddSingleUser(string application, string identfier, DateTime time)
        {
            lock (this.lockObject)
            {
                this.singleUsers[identfier] = application;
                this.SaveToFile();
            }
        }

        /// <summary>
        /// Adds a feed aggregator.
        /// </summary>
        /// <param name="application">The application.</param>
        /// <param name="subscribers">The number of subscribers.</param>
        /// <param name="time">The time.</param>
        public void AddFeedAggregator(string application, int subscribers, DateTime time)
        {
            lock (this.lockObject)
            {
                this.feedAggregators[application] = subscribers;
                this.SaveToFile();
            }
        }

        /// <summary>
        /// Gets the statistcs which should be reported.
        /// </summary>
        /// <returns>The statistcs which should be reported.</returns>
        public FeedStatistics GetStatistcsToReport()
        {
            var statistics = new FeedStatistics();

            lock (this.lockObject)
            {
                statistics.TotalReaders = this.singleUsers.Count + this.feedAggregators.Values.Sum(v => v);

                foreach (var application in this.feedAggregators)
                {
                    statistics.ReadersByApplication[application.Key] = application.Value;
                }

                foreach (var application in this.singleUsers.Values.Distinct())
                {
                    statistics.ReadersByApplication[application] = this.singleUsers.Count(u => u.Value == application);
                }

                this.singleUsers.Clear();
                this.feedAggregators.Clear();

                this.currentDay = DateTime.Now.Day;
            }

            return statistics;
        }

        /// <summary>
        /// Reads the statistics from a file to fill the cache.
        /// </summary>
        private void ReadFromFile()
        {
            if (File.Exists(this.filename))
            {
                var document = XDocument.Load(this.filename);

                this.currentDay = int.Parse(document.Root.Attribute("currentday").Value);

                foreach (var user in document.Root.Element("singleusers").Descendants())
                {
                    this.singleUsers.Add(user.Attribute("id").Value, user.Attribute("app").Value);
                }

                foreach (var user in document.Root.Element("aggregators").Descendants())
                {
                    this.feedAggregators.Add(user.Attribute("app").Value, int.Parse(user.Attribute("subscribers").Value));
                }
            }
            else
            {
                this.currentDay = DateTime.Now.Day;
            }
        }

        /// <summary>
        /// Saves the statistics to a file.
        /// </summary>
        private void SaveToFile()
        {
            var users = new XElement("singleusers");

            foreach (var user in this.singleUsers)
	        {
                users.Add(new XElement(
                    "user",
                    new XAttribute("id", user.Key),
                    new XAttribute("app", user.Value)));
	        }

            var aggregators = new XElement("aggregators");

            foreach (var aggregator in this.feedAggregators)
            {
                aggregators.Add(new XElement(
                    "aggregator",
                    new XAttribute("app", aggregator.Key),
                    new XAttribute("subscribers", aggregator.Value)));
            }

            var document = new XDocument();
            document.Declaration = new XDeclaration("1.0", "UTF-8", "yes");
            document.Add(new XElement(
                "feedstatistics",
                new XAttribute("currentday", this.currentDay),
                users,
                aggregators));

            document.Save(this.filename);
        }
    }
}